from setuptools import setup

setup(
    name='reem',
    version='v0.0.1',
    packages=['reem'],
    url='https://www.github.com/tn74/reem',
    license='MIT',
    author='Trishul Nagenalli',
    author_email='tn74@duke.edu',
    description='Redis Extendable Efficient Middleware'
)
